# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionState.Login do
  require Amethyst.ConnectionState.Macros
  alias Amethyst.ConnectionState.Macros

  require Logger

  @moduledoc """
  This module contains the packets and logic for the Login state.
  """
  Macros.defpacket_clientbound :disconnect, 0x00, 767, [reason: :json]
  Macros.defpacket_clientbound :encryption_request, 0x01, 767, [
    server_id: :string,
    public_key: :byte_array,
    verify_token: :byte_array,
    should_authenticate: :bool
  ]
  Macros.defpacket_clientbound :login_success, 0x02, 767, [
    uuid: :uuid,
    username: :string,
    properties: {:array, [
      name: :string,
      value: :string,
      signature: {:optional, :string}
    ]},
    strict_error_handling: :bool
  ]
  Macros.defpacket_clientbound :set_compression, 0x03, 767, [threshold: :varint]
  Macros.defpacket_clientbound :login_plugin_request, 0x04, 767, [
    message_id: :varint,
    channel: :string,
    data: :raw
  ]
  Macros.defpacket_clientbound :cookie_request, 0x05, 767, [identifier: :string]

  Macros.defpacket_serverbound :login_start, 0x00, 767, [name: :string, player_uuid: :uuid]
  Macros.defpacket_serverbound :encryption_response, 0x01, 767, [
    shared_secret: :byte_array,
    verify_token: :byte_array
  ]
  Macros.defpacket_serverbound :login_plugin_response, 0x02, 767, [
    message_id: :varint,
    data: {:optional, :raw}
  ]
  Macros.defpacket_serverbound :login_acknowledged, 0x03, 767, []
  Macros.defpacket_serverbound :cookie_response, 0x04, 767, [identifier: :string, payload: {:optional, :byte_array}]

  def handle(%{packet_type: :login_start, name: name, player_uuid: player_uuid}, 767, state) do
    Logger.debug("Received login start for #{name} with UUID #{player_uuid}")
    if Application.get_env(:amethyst, :encryption, true) do
      verify_token = :crypto.strong_rand_bytes(4)
      public_key = Amethyst.Keys.get_pub()
      Logger.debug("Public key: #{inspect(public_key, limit: :infinity)}")
      send(self(), {:send_packet, %{
        packet_type: :encryption_request,
        server_id: "",
        public_key: public_key,
        verify_token: verify_token,
        should_authenticate: Application.get_env(:amethyst, :auth, false)
      }})
      state |> Map.put(:verify_token, verify_token) |> Map.put(:name, name) |> Map.put(:uuid, player_uuid)
    else
      send(self(), {:send_packet, %{
        packet_type: :login_success,
        uuid: player_uuid,
        username: name,
        properties: [],
        strict_error_handling: true
      }})
      :ok
    end
  end
  def handle(%{packet_type: :encryption_response, shared_secret: secret, verify_token: verify_token}, 767, state) do
    secret = Amethyst.Keys.decrypt(secret)
    verify_token = Amethyst.Keys.decrypt(verify_token)
    if verify_token == Map.get(state, :verify_token, :never) do
      send(self(), {:set_encryption, secret})

      if Application.get_env(:amethyst, :compression, nil) != nil do
        threshold = Application.get_env(:amethyst, :compression, 0)
        send(self(), {:send_packet, %{packet_type: :set_compression, threshold: threshold}})
        send(self(), {:set_compression, threshold})
      end

      if Application.get_env(:amethyst, :auth, false) == false do
        # Don't check authentication
        send(self(), {:send_packet, %{
          packet_type: :login_success,
          uuid: Map.get(state, :uuid),
          username: Map.get(state, :name),
          properties: [],
          strict_error_handling: true
        }})
        Map.put(state, :authenticated, false)
      else
        # Check authentication
        pubkey = Amethyst.Keys.get_pub()
        hash = Amethyst.Minecraft.Sha1.hash(secret <> pubkey)
        url = Application.get_env(:amethyst, :session_server, "https://sessionserver.mojang.com") <> "/session/minecraft/hasJoined?username=" <>
          Map.get(state, :name) <> "&serverId=" <> hash # I don't think we need to verify the IP in the use case of Amethyst...

        response = Req.get!(url,
        headers: [
          {"user-agent", "Amethyst/1.0"}
        ]).body

        <<c1::binary-size(8), c2::binary-size(4), c3::binary-size(4), c4::binary-size(4), c5::binary>> = response["id"]
        uuid = [c1, c2, c3, c4, c5] |> Enum.join("-")

        send(self(), {:send_packet, %{
          packet_type: :login_success,
          uuid: uuid,
          username: response["name"],
          properties: response["properties"] |>
            Enum.map(fn prop -> %{name: prop["name"], value: prop["value"], signature: Map.get(prop, "signature")} end),
          strict_error_handling: true
        }})
        Map.put(state, :authenticated, true) |> Map.put(:uuid, uuid) |> Map.put(:name, response["name"]) |> Map.put(:properties, response["properties"])
      end
    else
      raise RuntimeError, "Invalid verify token. Broken encryption?"
    end
  end
  def handle(%{packet_type: :login_acknowledged}, 767, _state) do
    Logger.debug("Received login acknowledged")
    send(self(), {:set_state, Amethyst.ConnectionState.Configuration})
    :ok
  end

  def disconnect(reason) do
    %{packet_type: :disconnect, reason:
      %{
        "text" => reason
      }
    }
  end
end
